{% extends "base.txt.j2" %}
{% block content %}
Tag Name: {{ data.tag|trim }}

Added tags:

{%- for tag in data.added_tags %}
    {{ tag|pprint }}
{%- endfor%}

Removed tags:

{%- for tag in data.removed_tags %}
    {{ tag|pprint }}
{%- endfor%}

Module:

{{ data.module|pprint }}

{% if data.build_url %}
Check for more details: {{ data.build_url|trim }}
{% endif %}

{% endblock %}
